<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: de0846735ec7 $
 * $Author: Lukas Imhof $
 * $Date: 2021-03-04 13:32 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
abstract class we_base_country{
	public const TERRITORY = 'territory';
	public const REGION = 'region';
	public const LANGUAGE = 'language';
	public const SCRIPT = 'script';
	public const MONTH = 'months';
	public const DAY = 'days';
	public const LANGDIR = WE_INCLUDES_PATH . 'language/';

	/** @var array<string,array> */
	private static $langCache = [];

	public static function getTranslation(string $countrykey, string $type, string $langcode): string{
		[$langcode] = explode('_', strtolower($langcode), 2);
		$class = 'we_country_' . $langcode;
		return empty($class::DATA[$type][$countrykey]) ? '' : $class::DATA[$type][$countrykey];
	}

	public static function getTranslationList(string $type, string $langcode): array{
		[$langcode] = explode('_', strtolower($langcode), 2);
		$class = 'we_country_' . $langcode;
		return empty($class::DATA[$type]) ? [] : $class::DATA[$type];
	}

	public static function dateformat(string $langcode, DateTime $date, string $format): string{
		$months = self::getTranslationList(self::MONTH, $langcode);
		$days = self::getTranslationList(self::DAY, $langcode);

		$dat = $date->format(strtr($format, [
			'D' => "\dD\d", //Mon bis Sun
			'l' => "\lD\l", //Sunday bis Saturday
			'F' => "\\fn\\f", //January bis December
			'M' => '\mn\m', //Jan bis Dec
		]));

		$wd = $date->format('D');
		$mon = $date->format('n');

		return strtr($dat, [
			'd' . $wd . 'd' => $days['abbreviated'][strtolower($wd)], //Mon bis Sun
			'l' . $wd . 'l' => $days['wide'][strtolower($wd)], //Sunday bis Saturday
			'f' . $mon . 'f' => $months['wide'][$mon], //January bis December
			'm' . $mon . 'm' => $months['abbreviated'][$mon], //Jan bis Dec
			]
		);
	}

	/**
	 *
	 * @return array<string,string>
	 */
	public static function getWELangs(): array{
		return [
			'de' => 'Deutsch',
			'en' => 'English',
			'es' => 'Spanish',
			'fi' => 'Finnish',
			'fr' => 'French',
			'nl' => 'Dutch',
			'pl' => 'Polish',
			'ru' => 'Russian',
		];
	}

	/**
	 *
	 * @return array<string,string>
	 */
	public static function getWECountries(): array{
		return [
			'DE' => 'de',
			'GB' => 'en',
			'NL' => 'nl',
			'FI' => 'fi',
			'RU' => 'ru',
			'ES' => 'es',
			'PL' => 'pl',
			'FR' => 'fr'
		];
	}

	/**
	 * @return array<string,string>
	 */
	public static function getWeFrontendLanguagesForBackend(): array{
		$la = [];
		if(empty(we_base_WE::$Langs['FrontendLanguages'])){
			return [];
		}
		$targetLang = array_search(we_base_WE::$Language, self::getWELangs(), true) ?: 'de';
		foreach(we_base_WE::$Langs['FrontendLanguages'] as $Locale){
			$temp = explode('_', $Locale);
			$la[$Locale] = (count($temp) == 1 ?
				we_base_country::getTranslation($temp[0], we_base_country::LANGUAGE, $targetLang) . ' ' . $Locale :
				we_base_country::getTranslation($temp[0], we_base_country::LANGUAGE, $targetLang) . ' (' . we_base_country::getTranslation($temp[1], we_base_country::TERRITORY, $targetLang) . ') ' . $Locale);
		}
		return $la;
	}

	private static function getVarArray(string $name, string $string)/* :mixed */{
		$return = self::$langCache[$name];
		$arr_matches = [];
		preg_match_all('/\[([^\]]*)\]/', $string, $arr_matches, PREG_PATTERN_ORDER);
//		t_e($return, $arr_matches);
		foreach($arr_matches[1] as $dimension){
			if(!isset($return[$dimension])){
				throw new we_exception_base('Requested lang entry l_' . $name . $string . ' not found!');
			}
			$return = $return[$dimension];
		}
		return $return;
	}

	private static function setCache(string $name, string $specific): void{
		if(!empty(self::$langCache[$name])){
			return;
		}
		$dirs = (empty($_SESSION['weS']['gl']) ? [] : $_SESSION['weS']['gl']);
		$dirs[] = self::LANGDIR;
		foreach($dirs as $dir){
			$file = $dir . we_base_WE::$Language . '/' . str_replace('_', '/', $name) . '.inc.php';
			if(file_exists($file)){
				$data = include($file);
				if($data){
					self::$langCache[$name] = $data;
				}
				return;
			}
		}
		throw new we_exception_file('Language file "' . $name . '" not found with entry ' . $specific);
	}

	/**
	 *
	 * @param string $name
	 * @param string $specific
	 * @return string[]
	 */
	public static function g_l_a(string $name, string $specific): array{
		try{
			self::setCache($name, $specific);
			return self::getVarArray($name, $specific);
		} catch (Throwable $e){
			t_e($e->getMessage());
			return [];
		}
	}

	/**
	 *
	 * @param string $name
	 * @param string $specific
	 * @param bool $omitErrors
	 * @param bool $forHtml
	 * @return string
	 * @throws we_exception_base
	 */
	public static function g_l(string $name, string $specific, bool $omitErrors = false, bool $forHtml = false): string{
		try{
			self::setCache($name, $specific);
			$ret = (string) self::getVarArray($name, $specific);
			return ($forHtml ? strtr(htmlspecialchars($ret), ["\n" => we_html_element::br()]) : $ret);
		} catch (Throwable $e){
			if($omitErrors){
				throw new we_exception_base('Lang: entry for language "' . we_base_WE::$Language . '" not found ' . $name . $specific);
			}
			t_e($e->getMessage());
			return $name . $specific;
		}
	}

}
