<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: cdc2c0325715 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2021-12-02 22:36 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_base_request{
	/** @var array<string,string> */
	private static $allTables = [];

	public const TYPE_DEFAULT = '__TYPE_DEFAULT__'; // to be used as default just for indicating that is IS an default
	public const NOT_VALID_S = '__NOT_VALID_S__';
	public const NOT_VALID_AS = ['__NOT_VALID_S__'];
	public const NOT_VALID_RAW = '__NOT_VALID_RAW__';
	public const NOT_VALID_I = PHP_INT_MAX;
	public const NOT_VALID_AI = [PHP_INT_MAX];

	/* converts an csv of ints to an array */
	public const INTLISTA = 'intListA';

	/** convert all to real integers */
	public const INT = 'int';

	/** convert all to real floats */
	public const FLOAT = 'float';

	/** convert to true/false */
	public const BOOL = 'bool';

	/** don't change the string, e.g. for html-code */
	public const RAW = 'raw';

	/** unserialize the give string */
	public const JSON = 'json';

	/** check if the given string is a correct url, must match prot://user:pwd@domain/path */
	public const URL = 'url';
	public const URL_PARAM = 'url_param';

	/** checks if this is a correct email-adress */
	public const EMAIL = 'email'; //add email_list
	public const EMAIL_FROM = 'email_from'; // temporary until email validation pattern is fixed

	/** strip all (html-)tags */
	public const STRING = 'string';

	/** unused, must be fixed */
	//FIXME: implement
	public const DATE = 'date';
	//FIXME: implement
	public const DATE_TIME = 'datetime';
	public const UNIX_TIME = 'unixtime';

	/** encode html for inner-tag-transport, encodes HTML entities */
	public const HTML = 'html';

	/** @var string remove/escape JS and PHP code */
	public const WYSIWYG = 'wysiwyg';
	public const TRANSACTION = 'transaction';

	/** @deprecated 	 */
	public const INTLIST = 'intList';
	public const UNIT = 'unit';
	public const TABLE = 'table';
	public const FILE = 'file';
	public const FILELIST = 'filelist';
	public const FILELISTA = 'filelista';
	public const EXTFILE = 'extfile';
	public const STRING_LIST = 'stringL';
	public const EMAILLIST = 'emailL';
	public const EMAILLISTA = 'emailLA';
	public const FILENAME = 'filename';
	// the following types do not sanitize, so they will allow spaces,...
	public const WEFILE = 'wefile';
	public const WEFILELIST = 'wefilelist';
	public const WEFILELISTA = 'wefilelista';
//remove these types!
	public const JS = 'js';
	public const SERIALIZED = 'serial';
	public const SERIALIZED_JSON = 'serialJ';
	public const SERIALIZED_KEEP = 'serialK';
//only temporary
	public const RAW_CHECKED = 'rawC';

	/** Helper for Filtering variables (callback of array_walk)
	 *
	 * @param string|array<string,string> $var value
	 * @param string $key key used by array-walk - unused
	 * @param array{string,mixed} $data array pair of type & default
	 * @return void
	 */
	private static function _weRequest(&$var, string $key, array $data): void{//@phan-suppress-current-line PhanUnusedPrivateMethodParameter
		if(is_array($var)){
			array_walk($var, 'we_base_request::_weRequest', $data);
			return;
		}

		[$type, $default] = $data;
		switch($type){
			case self::TRANSACTION:
				$var = (preg_match('|^([a-f0-9]){32}$|i', (string) $var) ? (string) $var : $default);
				return;
			case self::INTLISTA:
				$var = trim((string) $var, ',');
				$var = $var || $var === '0' ? array_map('intval', explode(',', $var)) : $default;
				return;
			case self::INTLIST://@phan-suppress-current-line PhanDeprecatedClassConstant
				$var = trim((string) $var, ',');
				$var = $var || $var === '0' ? implode(',', array_map('intval', explode(',', $var))) : $default;
				return;
			case self::SERIALIZED:
			case self::SERIALIZED_JSON:
				$var = we_unserialize((string) $var);
				return;
			case self::SERIALIZED_KEEP:
				return;
			case self::UNIT:
				$regs = []; //FIMXE: check for %d[em,ex,pt,%...]?
				$var = (preg_match('/(\d+\.?\d*) ?(em|ex|pt|px|in|mm|cm|pc|ch|rem|vw|vh|vmin|vmax|%)?/', (string) $var, $regs) ? $regs[1] . ($regs[2] ?? '') : '' );
				return;
			case self::INT:
				$var = (int) ((string) $var === '' ? $default : $var);
				return;
			case self::FLOAT:
				//FIXME: check for country dependencies (eg. 1.3333,22)
				$var = (float) ((string) $var === '' ? $default : we_base_util::std_numberformat((string) $var));
				return;
			case self::BOOL:
				if(is_bool($var)){
					return;
				}
				$var = self::convertBool((string) $var);
				return;
			case self::TABLE: //FIXME: this doesn't hold for OBJECT_X_TABLE - make sure we don't use them in requests
				$var = ($var && in_array((string) $var, self::$allTables, true)) ? (string) $var : $default;
				return;
			case self::EMAILLISTA:
			case self::EMAILLIST:
				$mails = self::convertMail((string) $var);
				$var = ($type === self::EMAILLISTA ? $mails : implode(',', $mails));
				return;
			case self::EMAIL://removes mailto:
			case self::EMAIL_FROM:
				$mails = self::convertMail((string) $var, true, $type === self::EMAIL_FROM);
				$var = empty($mails) ? $default : $mails[0];
				return;
			case self::WEFILELIST:
			case self::WEFILELISTA:
			case self::FILELISTA:
			case self::FILELIST:
				$var = explode(',', trim(strtr((string) $var, ['../' => '', './' => '/', '..' => '', '//' => '/']), ','));
				foreach($var as &$cur){
					$cur = (string) ($type === self::WEFILELIST || $type === self::WEFILELISTA ? $cur : filter_var($cur, FILTER_SANITIZE_URL));
					if($cur === rtrim(WEBEDITION_DIR, '/') || str_starts_with($cur, WEBEDITION_DIR)){//file-selector has propably access
						if(!(str_contains($cur, WE_SITE_DIR) || str_contains($cur, TEMP_DIR))){//allow site/tmp dir
							$cur = we_base_WE::isDebug() ? $cur : '-1';
						}
					}
				}
				$var = array_filter($var);
				$var = ($type === self::FILELIST || $type === self::WEFILELIST ? implode(',', $var) : $var);
				return;
			case self::WEFILE:
			case self::FILE:
			case self::EXTFILE:
				$var = strtr(($type === self::WEFILE ? (string) $var : (string) filter_var((string) $var, FILTER_SANITIZE_URL)), ['../' => '', './' => '/', '..' => '', '//' => '/']);
				if(str_starts_with($var, rtrim(WEBEDITION_DIR, '/'))){//file-selector has propably access
					if($type === self::EXTFILE){
						$var = '';
					} elseif(!(str_contains($var, WE_SITE_DIR) || str_contains($var, TEMP_DIR))){//allow site/tmp dir
						$var = we_base_WE::isDebug() ? $var : '-1';
					}
				}
				$var = $var ?: $default;
				return;
			case self::URL:
				$var = self::convertURL((string) $var);
				return;
			case self::FILENAME:
				$var = strtr($var, ['./' => '', '/' => '', '\\' => '']); // FIXME: use we_base_constants::CHARS_FILE
			/* falls through */
			case self::STRING://strips tags
				$var = filter_var((string) $var, FILTER_SANITIZE_STRING, FILTER_FLAG_NO_ENCODE_QUOTES);
				return;
			case self::STRING_LIST:
//in lists, we don't accept quotes
				$var = array_filter(array_map('trim', explode(',', filter_var((string) $var, FILTER_SANITIZE_STRING) ?: '')));
				return;
			case self::HTML://change html code to encoded values
				$var = filter_var(htmlspecialchars_decode((string) $var), FILTER_SANITIZE_SPECIAL_CHARS);
				return;
			case self::WYSIWYG:
				$var = strtr((string) $var, ['<?' => '<!--?', '?>' => '?-->', '<script' => '<!--script', '</script>' => '</script-->']);
				return;
			case self::JSON:
				$var = we_unserialize(base64_decode((string) $var) ?: '');
				return;
			/* case self::UNIX_TIME:
			  $var = $var ? new DateTime('@' . ((string) $var)) : null;
			  return; */
			case self::DATE:
			case self::DATE_TIME:
				try{
					$var = $var ? new DateTime((string) $var) : null;
				} catch (Throwable $e){
					$var = new DateTime();
				}
				return;
			case self::JS://for information!
			case self::RAW:
			case self::RAW_CHECKED:
//do nothing - used as placeholder for all types not yet known
				settype($var, 'string'); //make sure we have a string
				return;
			default:
				t_e('unknown filter type ' . $type);
		}
		$var = $default;
	}

	public static function convertBool(string $var): bool{
		switch(strtolower(trim($var))){
			case 'off':
			case 'false':
				return false;
			case 'true':
			case 'on':
				return true;
			default:
				return (bool) $var;
		}
	}

	private static function convertURL(string $var): string{
		if($var === '' || $var === '/'){
			return $var;
		}
		// prefix / or // to full qualified urls, so we can test, but keep the original
		$test = ($var[0] === '/' ?
			'http://' . ($var[1] !== '/' ? 'example.org' : '') . $var :
			$var);

		//full qualified url
		if(!preg_match('_^(?:(?:[a-z]+)://|/)(?:\S+(?::\S*)?@)?(?:(?!10(?:\.\d{1,3}){3})(?!127(?:\.\d{1,3}){3})(?!169\.254(?:\.\d{1,3}){2})(?!192\.168(?:\.\d{1,3}){2})(?!172\.(?:1[6-9]|2\d|3[0-1])(?:\.\d{1,3}){2})(?:[1-9]\d?|1\d\d|2[01]\d|22[0-3])(?:\.(?:1?\d{1,2}|2[0-4]\d|25[0-5])){2}(?:\.(?:[1-9]\d?|1\d\d|2[0-4]\d|25[0-4]))|(?:(?:[a-z\x{00a1}-\x{ffff}0-9]+-?)*[a-z\x{00a1}-\x{ffff}0-9]+)(?:\.(?:[a-z\x{00a1}-\x{ffff}0-9]+-?)*[a-z\x{00a1}-\x{ffff}0-9]+)*(?:\.(?:[a-z\x{00a1}-\x{ffff}]{2,})))(?::\d{2,5})?(?:/[^\s]*)?$_iuS', $test)){

			$tmp = explode(':', $var, 2);
			[$prot, $path] = (count($tmp) == 2 ? $tmp : ['file', $var]);

			if(preg_match('-(' . we_base_link::TYPE_INT_PREFIX . '|' . we_base_link::TYPE_MAIL_PREFIX . '|' . we_base_link::TYPE_OBJ_PREFIX . '|' . we_base_link::TYPE_THUMB_PREFIX . ')-', $var)){
				return $var;
			}
			switch($prot . ':'){
				case we_base_link::TYPE_INT_PREFIX:
				case we_base_link::TYPE_MAIL_PREFIX:
				case we_base_link::TYPE_OBJ_PREFIX:
					//FIXME: check types here too
					return $var;
				case we_base_link::TYPE_THUMB_PREFIX:
					if(!preg_match('|(\d+,\d+)|', $path)){
						return '';
					}
					return $var;
				case 'http:':
				case 'https:':
					return $var;
			}

			return $var;
		}
		return $var;
	}

	/**
	 *
	 * @param string $var
	 * @param bool $single
	 * @param bool $onlyFrom
	 * @return string[]
	 */
	private static function convertMail(string $var, bool $single = false, $onlyFrom = false): array{
//FIXME we need to improve this for "test, x" <a@b.de>, "test2, x" <b@d.de>
//preg_match_all('/("[\S\s]+"\s*|[^,"<>@]*\s+|)<?([^@<>,]*)@([a-zA-Z\d.-]+)>?/', $mail, $regs,PREG_SET_ORDER);
		$var = str_replace(we_base_link::TYPE_MAIL_PREFIX, '', $var);
		$mails = array_map('trim', ($single ?
			[$var] :
			explode(',', $var)
		));
		$regs = [];

		// temporary until email validation pattern is fixed
		$patternFrom = $onlyFrom ? '^[^@<>§%$&?"\':;*#=]+|' : '';

		foreach($mails as &$mail){
			if(!preg_match('-(["\'][\S\s]+["\']\s*|' . $patternFrom . '\S+\s*)<(\S+)@(\S+)>-', $mail, $regs)){ //mail formats "yy" <...@...>, =..... <...@...>
				//if format didn't match, filter the whole var as one address
				$regs = array_merge(['', ''], explode('@', trim($mail, '<>'), 2));
				if(!isset($regs[3])){
					$mail = '';
					continue;
				}
			}
			$host = (function_exists('idn_to_ascii') && defined('IDNA_DEFAULT') && defined('INTL_IDNA_VARIANT_UTS46') ? idn_to_ascii($regs[3], IDNA_DEFAULT, INTL_IDNA_VARIANT_UTS46) : $regs[3]);
			$mail = (filter_var($regs[2] . '@' . $host, FILTER_VALIDATE_EMAIL) !== false ?
				$regs[1] . ($regs[1] ? '<' : '') . $regs[2] . '@' . $regs[3] . ($regs[1] ? '>' : '') :
				'');
		}//if format didn't match, filter the whole var as one address

		return array_filter($mails);
	}

	private static function getTypeDefault(string $type, /* mixed */ $default = self::TYPE_DEFAULT)/* :mixed */{
		if($default !== self::TYPE_DEFAULT){
			return $default;
		}
//make sure we get the correct type instead
		switch($type){
			case self::BOOL:
				return false;
			case self::INT:
				return 0;
			case self::FLOAT:
				return 0.0;
			case self::EMAIL:
			case self::FILE:
			case self::FILELIST:
			case self::FILENAME:
			case self::HTML:
			case self::INTLIST://@phan-suppress-current-line PhanDeprecatedClassConstant
			case self::RAW:
			case self::RAW_CHECKED:
			case self::SERIALIZED_KEEP:
			case self::STRING:
			case self::TABLE:
			case self::TRANSACTION:
			case self::UNIT:
			case self::URL:
			case self::WEFILE:
			case self::WEFILELIST:
				return '';
			case self::FILELISTA:
			case self::INTLISTA:
			case self::JSON:
			case self::SERIALIZED:
			case self::SERIALIZED_JSON:
			case self::STRING_LIST:
			case self::WEFILELISTA:
				return [];
			case self::DATE:
			case self::DATE_TIME:
				return new DateTime();
		}
		throw new we_exception_base('Default type not implemented ' . $type);
	}

	private static function has_string_keys(array $array): bool{
		return count(array_filter(array_keys($array), 'is_string')) > 0;
	}

	/**
	 * @param mixed $var
	 * @param string $varType
	 * @param mixed $default
	 * @return mixed
	 * @deprecated since version 8.1
	 */
	public static function filterVar(/* mixed */$var, string $varType, /* mixed */ $default = self::TYPE_DEFAULT)/* :mixed */{
		self::_weRequest($var, '', [$varType, self::getTypeDefault($varType, $default)]);
		return $var;
	}

	/**
	 * Filter an Requested variable
	 * Note: every parameter after default is an optional index
	 * @param string $type type to filter, see list in _weGetVar
	 * @param string $name name of variable in Request array
	 * @param mixed $default default value
	 * @param int|string ...$args
	 * @return mixed default, if value not set, the filtered value else
	 */
	public static function _(string $type, string $name, /* mixed */ $default = self::TYPE_DEFAULT, ...$args)/* :mixed */{
		if(!isset($_REQUEST)){
			return $default;
		}
		$var = $_REQUEST;
		array_unshift($args, $name);

		foreach($args as $arg){
			if(is_string($var) || !is_array($var) || !isset($var[$arg])){
				return self::getTypeDefault($type, $default);
			}
			$var = $var[$arg];
		}

		if(is_array($var)){
			array_walk($var, 'we_base_request::_weRequest', [$type, self::getTypeDefault($type, $default), false]);
			$var = self::has_string_keys($var) || in_array($type, [self::BOOL, self::INT, self::RAW, self::RAW_CHECKED]) ? $var : array_filter($var);
			return $var ?: self::getTypeDefault($type, $default);
		}
		self::_weRequest($var, '', [$type, self::getTypeDefault($type, $default)]);
		return $var;
	}

	/**
	 * @param string $name
	 * @param int|string ...$args
	 * @return mixed
	 * @throws we_exception_base
	 */
	private static function getVar(string $name, ...$args){
		if(!isset($_REQUEST)){
			throw new we_exception_base();
		}
		$var = $_REQUEST;
		array_unshift($args, $name);

		foreach($args as $arg){
			if(is_string($var) || !is_array($var) || !isset($var[$arg])){
				throw new we_exception_base();
			}
			$var = $var[$arg];
		}
		return $var;
	}

	public static function bool(string $name, bool $default = false, ...$args): bool{
		try{
			$var = self::getVar($name, ...$args);
			self::_weRequest($var, '', [self::BOOL, $default]);
			return (bool) $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 * @param string $name
	 * @param bool[] $default
	 * @param int|string ...$args
	 * @return bool[]
	 */
	public static function boolList(string $name, array $default = [], ...$args): array{
		try{
			$var = self::getVar($name, ...$args);
			if(is_array($var)){
				self::_weRequest($var, '', [self::BOOL, $default]);
			} else {
				// FIXME: implement in function intList
				t_e('we_base_request::boolList: string given instead of array', $var);
			}
			assert(is_array($var));
			'@phan-var bool[] $var';
			return $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 *
	 * @param string $name
	 * @param int $default
	 * @param int|string ...$args
	 * @return int
	 */
	public static function int(string $name, int $default = 0, ...$args): int{
		try{
			$var = self::getVar($name, ...$args);
			self::_weRequest($var, '', [self::INT, $default]);
			return (int) $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 * @param string $name
	 * @param int[] $default
	 * @param int|string ...$args
	 * @return int[]
	 */
	public static function intList(string $name, array $default = [], ...$args): array{
		try{
			$var = self::getVar($name, ...$args);
			if(is_array($var)){
				self::_weRequest($var, '', [self::INT, 0]);
				$var = array_filter((array) $var);
			} else {
				self::_weRequest($var, '', [self::INTLISTA, $default]);
			}
			assert(is_array($var));
			'@phan-var int[] $var';
			return $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 *
	 * @param string $name
	 * @param float $default
	 * @param int|string ...$args
	 * @return float
	 */
	public static function float(string $name, float $default = 0, ...$args): float{
		try{
			$var = self::getVar($name, ...$args);
			self::_weRequest($var, '', [self::FLOAT, $default]);
			return (float) $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 *
	 * @param string $name
	 * @param string $default
	 * @param int|string ...$args
	 * @return string
	 */
	public static function string(string $name, string $default = '', ...$args): string{
		try{
			$var = self::getVar($name, ...$args);

			self::_weRequest($var, '', [self::STRING, $default]);
			return (string) $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 * @param string $name
	 * @param string[] $default
	 * @param int|string ...$args
	 * @return string[]
	 */
	public static function stringList(string $name, array $default = [], ...$args): array{
		try{
			$var = self::getVar($name, ...$args);
			if(is_array($var)){
				self::_weRequest($var, '', [self::STRING, '']);
				$var = array_filter((array) $var);
			} else {
				self::_weRequest($var, '', [self::STRING_LIST, $default]);
			}
			assert(is_array($var));
			'@phan-var string[] $var';
			return $var;
		} catch (Exception $ex){
			return $default;
		}
	}

	/**
	 *
	 * @param string $name
	 * @param null|string|DateTime $default
	 * @param int|string ...$args
	 * @return string
	 */
	public static function date(string $name, /* ?string */ $default = 'now', ...$args): DateTime{
		try{
			$var = self::getVar($name, ...$args);
			self::_weRequest($var, '', [self::DATE, $default]);
			'@phan-var DateTime $var';
			if($var){
				return $var;
			}
			if($default === null){
				throw new we_exception_base('no date');
			}
			return ($default instanceof DateTime) ? $default : new DateTime($default);
		} catch (Exception $ex){
			if($default === null){
				throw new we_exception_base('no date');
			}
			return ($default instanceof DateTime) ? $default : new DateTime($default);
		}
	}

	/**
	 *
	 * @param mixed $var
	 * @param int[] $default
	 * @return int[]
	 */
	public static function filterBoolList(/* mixed */$var, array $default = []): array{
		if(is_array($var)){
			self::_weRequest($var, '', [self::BOOL, 0]);
			$var = array_filter((array) $var);
		} else {
			// FIXME: implement in function intList
			t_e('we_base_request::boolList: string given instead of array', $var);
		}
		assert(is_array($var));
		'@phan-var int[] $var';
		return $var;
	}

	/**
	 *
	 * @param mixed $var
	 * @param int[] $default
	 * @return int[]
	 */
	public static function filterIntList(/* mixed */$var, array $default = []): array{
		if(is_array($var)){
			self::_weRequest($var, '', [self::INT, 0]);
			$var = array_filter((array) $var);
		} else {
			self::_weRequest($var, '', [self::INTLISTA, $default]);
		}
		assert(is_array($var));
		'@phan-var int[] $var';
		return $var;
	}

	/**
	 *
	 * @param mixed $var
	 * @param string[] $default
	 * @return string[]
	 */
	public static function filterStringList(/* mixed */$var, array $default = []): array{
		if(is_array($var)){
			self::_weRequest($var, '', [self::STRING, 0]);
			$var = array_filter((array) $var);
		} else {
			self::_weRequest($var, '', [self::STRING_LIST, $default]);
		}
		assert(is_array($var));
		'@phan-var string[] $var';
		return $var;
	}

	/**
	 * @param array<string,string> $tables
	 */
	public static function registerTables(array $tables): void{
		self::$allTables = array_merge(self::$allTables, $tables);
	}

	/**
	 *
	 * @return array<string,string>
	 */
	public static function getAllTables(): array{
		return self::$allTables;
	}

	/* 	private static function unparse_url(array $parsed_url):string{
	  $user = $parsed_url['user'] ?? '';
	  $pass = isset($parsed_url['pass']) ? ':' . $parsed_url['pass'] : '';

	  return
	  (isset($parsed_url['scheme']) ? $parsed_url['scheme'] . '://' : '') .
	  (isset($parsed_url['host']) ? $parsed_url['host'] : '') .
	  (isset($parsed_url['port']) ? ':' . $parsed_url['port'] : '') .
	  ($user || $pass ? $pass . '@' : '') .
	  (isset($parsed_url['path']) ? str_replace(' ', '%20', $parsed_url['path']) : '') .
	  (isset($parsed_url['query']) ? '?' . $parsed_url['query'] : '') .
	  (isset($parsed_url['fragment']) ? '#' . $parsed_url['fragment'] : '');
	  } */

	/**
	 * this function closes any open request by the browser, but keeps running until killed by the server application
	 * @return void
	 */
	public static function finishRequest(): void{
		if(we_base_WE::$rebuildActive){
			return;
		}
		session_write_close();
		//flush();
		if(function_exists('fastcgi_finish_request')){
			fastcgi_finish_request();
		}
		ignore_user_abort(true);
	}

}
