<?php
declare(strict_types=1);

/**
 * Zend Framework
 *
 * LICENSE
 *
 * This source file is subject to the new BSD license that is bundled
 * with this package in the file LICENSE.txt.
 * It is also available through the world-wide-web at this URL:
 * http://framework.zend.com/license/new-bsd
 * If you did not receive a copy of the license and are unable to
 * obtain it through the world-wide-web, please send an email
 * to license@zend.com so we can send you a copy immediately.
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 * @version    $Id$
 */

/**
 * Smtp implementation of Mail_Protocol_Abstract
 *
 * Minimum implementation according to RFC2821: EHLO, MAIL FROM, RCPT TO, DATA, RSET, NOOP, QUIT
 *
 * @copyright  Copyright (c) 2005-2015 Zend Technologies USA Inc. (http://www.zend.com)
 * @license    http://framework.zend.com/license/new-bsd     New BSD License
 */
class we_mail_ProtocolSmtp extends we_mail_ProtocolAbstract{
	/** @var string The transport method for the socket	 */
	protected $transport = 'tcp';

	/** @var string Indicates that a session is requested to be secure	 */
	protected $secure;

	/** @var bool Indicates an smtp session has been started by the HELO command	 */
	protected $sess = false;

	/** @var bool Indicates the HELO command has been issues	 */
	protected $helo = false;

	/** @var bool Indicates an smtp AUTH has been issued and authenticated */
	protected $auth = false;

	/** @var bool Indicates a MAIL command has been issued	 */
	protected $mail = false;

	/** @var bool Indicates one or more RCTP commands have been issued	 */
	protected $rcpt = false;

	/** @var bool Indicates that DATA has been issued and sent */
	protected $data = false;

	/**
	 * Constructor.
	 *
	 * @param  string  $host
	 * @param  integer $port
	 * @param  array   $config
	 * @return void
	 * @throws we_exception_mail
	 */
	public function __construct(string $host = '127.0.0.1', int $port = 0, array $config = []){
		if(isset($config['ssl'])){
			switch(strtolower($config['ssl'])){
				case 'tls':
					$this->secure = 'tls';
					break;

				case 'ssl':
					$this->transport = 'ssl';
					$this->secure = 'ssl';
					$port = $port ?: 465;
					break;

				default:
					throw new we_exception_mail($config['ssl'] . ' is unsupported SSL type');
			}
		}

		// If no port has been specified then check the master PHP ini file. Defaults to 25 if the ini setting is null.
		$port = $port ?: ((int) ini_get('smtp_port')) ?: 25;
		parent::__construct($host, $port);
	}

	/**
	 * Connect to the server with the parameters given in the constructor.
	 *
	 * @return boolean
	 */
	public function connect(): bool{
		return $this->realConnect($this->transport . '://' . $this->host . ':' . $this->port);
	}

	/**
	 * Initiate HELO/EHLO sequence and set flag to indicate valid smtp session
	 *
	 * @param  string $host The client hostname or IP address (default: 127.0.0.1)
	 * @throws we_exception_mail
	 * @return void
	 */
	public function helo(string $host = '127.0.0.1'): void{
		if($this->socket === null){
			throw new we_exception_mail('socket not connected');
		}
// Respect RFC 2821 and disallow HELO attempts if session is already initiated.
		if($this->sess === true){
			throw new we_exception_mail('Cannot issue HELO to existing session');
		}

		// Initiate helo sequence
		$this->expect('220', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		$this->ehlo($host);

		// If a TLS session is required, commence negotiation
		if($this->secure == 'tls'){
			$this->send('STARTTLS');
			$this->expect('220', 180);
			if(!stream_socket_enable_crypto($this->socket, true, STREAM_CRYPTO_METHOD_SSLv23_CLIENT)){
				throw new we_exception_mail('Unable to connect via TLS');
			}
			$this->ehlo($host);
		}

		$this->startSession();
		$this->auth();
	}

	/**
	 * Send EHLO or HELO depending on capabilities of smtp host
	 *
	 * @param  string $host The client hostname or IP address (default: 127.0.0.1)
	 * @throws we_exception_mail
	 * @return void
	 */
	protected function ehlo(string $host): void{
		// Support for older, less-compliant remote servers. Tries multiple attempts of EHLO or HELO.
		try{
			$this->send('EHLO ' . $host);
			$this->expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		} catch (we_exception_mail $e){
			$this->send('HELO ' . $host);
			$this->expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		}
	}

	/**
	 * Issues MAIL command
	 *
	 * @param  string $from Sender mailbox
	 * @throws we_exception_mail
	 * @return void
	 */
	public function mail(string $from): void{
		if($this->sess !== true){
			throw new we_exception_mail('A valid session has not been started');
		}

		$this->send('MAIL FROM:<' . $from . '>');
		$this->expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		// Set mail to true, clear recipients and any existing data flags as per 4.1.1.2 of RFC 2821
		$this->mail = true;
		$this->rcpt = false;
		$this->data = false;
	}

	/**
	 * Issues RCPT command
	 *
	 * @param  string $to Receiver(s) mailbox
	 * @throws we_exception_mail
	 * @return void
	 */
	public function rcpt(string $to): void{
		if($this->mail !== true){
			throw new we_exception_mail('No sender reverse path has been supplied');
		}

		// Set rcpt to true, as per 4.1.1.3 of RFC 2821
		$this->send('RCPT TO:<' . $to . '>');
		$this->expect(['250', '251'], 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
		$this->rcpt = true;
	}

	/**
	 * Issues DATA command
	 *
	 * @param  string $data
	 * @throws we_exception_mail
	 * @return void
	 */
	public function data($data): void{
		// Ensure recipients have been set
		if($this->rcpt !== true){
			throw new we_exception_mail('No recipient forward path has been supplied');
		}

		$this->send('DATA');
		$this->expect('354', 120); // Timeout set for 2 minutes as per RFC 2821 4.5.3.2

		foreach(explode(we_mail_mime::LINEEND, $data) as $line){
			if(str_starts_with($line, '.')){
				// Escape lines prefixed with a '.'
				$line = '.' . $line;
			}
			$this->send($line);
		}

		$this->send('.');
		$this->expect('250', 600); // Timeout set for 10 minutes as per RFC 2821 4.5.3.2
		$this->data = true;
	}

	/**
	 * Issues the RSET command and validates answer
	 *
	 * Can be used to restore a clean smtp communication state when a transaction has been cancelled or commencing a new transaction.
	 *
	 * @return void
	 */
	public function rset(): void{
		$this->send('RSET');
		// MS ESMTP doesn't follow RFC, see [ZF-1377]
		$this->expect(['250', '220']);

		$this->mail = false;
		$this->rcpt = false;
		$this->data = false;
	}

	/**
	 * Issues the NOOP command and validates answer
	 *
	 * Not used by Mail, could be used to keep a connection alive or check if it is still open.
	 *
	 * @return void
	 */
	public function noop(): void{
		$this->send('NOOP');
		$this->expect('250', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
	}

	/**
	 * Issues the VRFY command and validates answer
	 *
	 * Not used by Mail.
	 *
	 * @param  string $user User Name or eMail to verify
	 * @return void
	 */
	public function vrfy(string $user): void{
		$this->send('VRFY ' . $user);
		$this->expect(['250', '251', '252'], 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
	}

	/**
	 * Issues the QUIT command and clears the current session
	 *
	 * @return void
	 */
	public function quit(): void{
		if($this->sess){
			$this->send('QUIT');
			$this->expect('221', 300); // Timeout set for 5 minutes as per RFC 2821 4.5.3.2
			$this->stopSession();
		}
	}

	/**
	 * Default authentication method
	 *
	 * This default method is implemented by AUTH adapters to properly authenticate to a remote host.
	 *
	 * @throws we_exception_mail
	 * @return void
	 */
	public function auth(): void{
		if($this->auth === true){
			throw new we_exception_mail('Already authenticated for this session');
		}
	}

	/**
	 * Start mail session
	 *
	 * @return void
	 */
	protected function startSession(): void{
		$this->sess = true;
	}

	/**
	 * Stop mail session
	 *
	 * @return void
	 */
	protected function stopSession(): void{
		$this->sess = false;
	}

}
