<?php
declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 6f3dab527777 $
 * $Author: Marc Kraemer <marc.kraemer@aldente-it.de> $
 * $Date: 2021-07-12 19:06 +0200 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

/**
 * General Definition of WebEdition Newsletter
 *
 */
class we_newsletter_newsletter extends we_newsletter_base{
	public const SAVE_PATH_NOK = -10;
	public const MALFORMED_SENDER = -1;
	public const MALFORMED_REPLY = -2;
	public const MALFORMED_TEST = -3;
	public const MALFORMED_GROUP = -4;
	const LOG_START_SEND = 'log_start_send';
	const LOG_RETRY = 'retry';
	const LOG_CONTINUE_SEND = 'log_continue_send';
	const LOG_END_SEND = 'log_end_send';
	const LOG_MAIL_SENT = 'mail_sent';
	const LOG_MAIL_FAILED = 'mail_failed';
	const LOG_MAIL_MALFORMED = 'email_malformed';
	const LOG_DOMAIN_NOK = 'domain_nok';
	const LOG_MAIL_BLACK = 'email_is_black';
	const LOG_CAMPAIGN_RESET = 'log_campaign_reset';
	const LOG_SAVE_NL = 'log_save_newsletter';
	const CSV_STATUS_ALL = 0;
	const CSV_STATUS_INVALID = 1;
	const CSV_STATUS_VALID = 2;
	const OP_EQ = 0;
	const OP_NEQ = 1;
	const OP_LE = 2;
	const OP_LEQ = 3;
	const OP_GE = 4;
	const OP_GEQ = 5;
	const OP_LIKE = 6;
	const OP_CONTAINS = 7;
	const OP_STARTS = 8;
	const OP_ENDS = 9;

	//properties
	/** @var int */
	var $ID = 0;

	/** @var int */
	var $ParentID = 0;

	/** @var bool */
	var $IsFolder = false;

	/** @var string */
	var $Text = '';

	/** @var string */
	var $Path = '/';

	/** @var string */
	var $Subject = '';

	/** @var string */
	var $Sender = '';

	/** @var string */
	var $Reply = '';

	/** @var string */
	var $Attachments = '';

	/** @var string */
	var $Customers = '';

	/** @var string */
	var $Emails = '';

	/** @var string */
	var $Test = '';

	/** @var int */
	var $Step = 0;

	/** @var int */
	var $Offset = 0;

	/** @var string[] */
	var $log = [];

	/** @var we_newsletter_block[] */
	var $blocks = [];

	/** @var we_newsletter_group[] */
	var $groups = [];

	/** @var bool */
	var $isEmbedImages = false;

	/** @var array<string,int> */
	protected $MediaLinks = [];

	/**
	 * Default Constructor
	 * Can load or create new Newsletter depends of parameter
	 *
	 * @param int $newsletterID
	 */
	function __construct(int $newsletterID = 0){
		parent::__construct(NEWSLETTER_TABLE);

		$this->persistent_slots = [
			'ID' => we_base_request::INT,
			'ParentID' => we_base_request::INT,
			'Text' => we_base_request::STRING,
			'Path' => we_base_request::FILE,
			'Subject' => we_base_request::STRING,
			'Sender' => we_base_request::EMAIL,
			'Reply' => we_base_request::EMAIL,
			'Test' => we_base_request::EMAIL,
			'Step' => we_base_request::INT,
			'Offset' => we_base_request::INT,
			'IsFolder' => we_base_request::BOOL,
			'isEmbedImages' => we_base_request::BOOL,
		];

		$this->addBlock();
		$this->addGroup();

		if($newsletterID){
			$this->load($newsletterID);
		}
	}

	/**
	 * get newsletter from database
	 *
	 * @param int $newsletterID
	 */
	function load(int $newsletterID = 0, bool $loadData = true): bool{
		$ret = parent::load($newsletterID);
		$this->Text = stripslashes($this->Text);
		$this->Path = ($this->Path ?: '/');
		$this->Subject = stripslashes($this->Subject);
		$this->groups = we_newsletter_group::getAllGroups($newsletterID);
		$this->blocks = we_newsletter_block::getAllBlocks($newsletterID);
		return $ret;
	}

	/**
	 * save newsletter in db
	 *
	 * @param bool $check
	 * @throws we_exception_newsletter
	 */
	function saveNewsletter(bool $check = true): void{
		//check addesses
		if($check && !$this->IsFolder){
			//throws on error
			$this->checkEmails();
		}

		if(!$this->checkParents($this->ParentID)){
			throw new we_exception_newsletter(g_l('modules_newsletter', '[path_nok]'), self::SAVE_PATH_NOK);
		}

		if($this->IsFolder){
			$this->fixChildsPaths();
		}

		if($this->Step != 0 || $this->Offset != 0){
			$this->addLog(we_newsletter_newsletter::LOG_CAMPAIGN_RESET);
		}
		$this->Step = 0;
		$this->Offset = 0;

		parent::save();

		if($this->IsFolder){
			return;
		}

		$this->db->query('DELETE FROM ' . NEWSLETTER_GROUP_TABLE . ' WHERE NewsletterID=' . intval($this->ID));
		$this->db->query('DELETE FROM ' . NEWSLETTER_BLOCK_TABLE . ' WHERE NewsletterID=' . intval($this->ID));

		foreach($this->groups as $group){
			$group->setNewsletterID($this->ID);
			$group->save();
		}
		$count_group = count($this->groups);
		foreach($this->blocks as $block){
			$groups = array_filter($block->getGroups());
			foreach($groups as $k => $v){
				if($v > $count_group){
					unset($groups[$k]);
				}
			}
			$block->setGroups($groups);
			$block->setNewsletterID($this->ID);
			$block->save();
			//#8199: thrown out all addslashes() and stripslashes() here!
		}

		$this->registerMediaLinks();
		$this->addLog(we_newsletter_newsletter::LOG_SAVE_NL);
	}

	function registerMediaLinks(): bool{
		$this->unregisterMediaLinks();
		$c = 1;
		foreach($this->blocks as $block){
			switch($block->Type){
				case 0:
				case 1:
					if($block->getLinkID()){
						$this->MediaLinks['Block ' . $c++ . ': [Dokument ID]'] = $block->getLinkID();
					}
					break;
				case 5:
					if($block->getHTML() && ($refs = we_wysiwyg_editor::getAllMediaLinks($block->getHTML(), 'NN'))){
						$c++;
						foreach($refs as $k => $v){
							$this->MediaLinks['Block ' . $c . ': ' . $k] = $v;
						}
					}
					break;
				case 6:
					if($block->getLinkID()){
						$this->MediaLinks['Block ' . $c++ . ': [Anhang ID]'] = $block->getLinkID();
					}
					break;
			}
		}

		// FIXME: inherit from we_base_model and call parent::registerMediaLinks();
		// filter MediaLinks by media contenttype
		if(!empty($this->MediaLinks)){
			$whereType = 'AND ContentType IN ("' . we_base_ContentTypes::APPLICATION . '","' . we_base_ContentTypes::IMAGE . '","' . we_base_ContentTypes::VIDEO . '","' . we_base_ContentTypes::PDF . '")';
			$verifiedIDs = $this->db->getAllq('SELECT ID FROM ' . FILE_TABLE . ' WHERE ID IN (' . implode(',', array_unique(array_values($this->MediaLinks))) . ') ' . $whereType, true);
		} else {
			$verifiedIDs = [];
		}
		$this->MediaLinks = array_intersect($this->MediaLinks, $verifiedIDs);

		if(empty($this->MediaLinks)){
			return true;
		}

		foreach($this->MediaLinks as $element => $remObj){
			$this->db->query('REPLACE INTO ' . FILELINK_TABLE . ' SET ' . we_database_base::arraySetter(['ID' => $this->ID,
					'DocumentTable' => stripTblPrefix($this->table),
					'type' => 'media',
					'remObj' => $remObj,
					'remTable' => stripTblPrefix(FILE_TABLE),
					'nHash' => is_numeric($element) ? null : sql_function('x\'' . md5($element) . '\''),
					'position' => 0,
					'isTemp' => 0
			]));
		}
		return true;
	}

	function unregisterMediaLinks(): bool{
		return $this->db->query('DELETE FROM ' . FILELINK_TABLE . ' WHERE ID=' . intval($this->ID) . ' AND DocumentTable="' . $this->db->escape(stripTblPrefix($this->table)) . '"  AND type="media"');
	}

	/**
	 * delete newsletter from database
	 *
	 * @return bool
	 */
	function delete(): bool{

		if($this->IsFolder){
			$this->deleteChilds();
		}
		foreach($this->blocks as &$block){
			$block->delete();
			$block = new we_newsletter_block();
		}
		foreach($this->groups as &$group){
			$group->delete();
			$group = new we_newsletter_group();
		}
		$this->clearLog();
		$this->unregisterMediaLinks();

		return parent::delete();
	}

	/**
	 * delete childs from database
	 *
	 */
	protected function deleteChilds(): bool{
		$ids = $this->db->getAllq('SELECT ID FROM ' . NEWSLETTER_TABLE . ' WHERE ParentID=' . intval($this->ID), true);

		$success = true;
		foreach($ids as $id){
			$child = new self((int) $id);
			$success = $success && $child->delete();
		}

		return $success;
	}

	/**
	 * gets all newsletter names from databes and returns them as an associated array with id as key and name as value
	 *
	 * @return array<int,string>
	 */
	function getAllNewsletter(): array{
		return DB_WE::$inst->getAllFirstq('SELECT ID,Text FROM ' . NEWSLETTER_TABLE . ' ORDER BY ID', false); //@phan-suppress-current-line PhanPartialTypeMismatchReturn
	}

	/**
	 * add a new block to a newsletter
	 *
	 * @param int $where
	 */
	function addBlock(int $where = -1): void{
		if($where == -1 || $where > count($this->blocks) - 1){
			$this->blocks[] = new we_newsletter_block();
			return;
		}
		$temp = [];
		foreach($this->blocks as $k => $v){
			if($k == $where){
				$temp[] = new we_newsletter_block();
			}
			$temp[] = $v;
		}
		$this->blocks = $temp;
	}

	/**
	 * remove newsletters block
	 *
	 * @param int $id
	 */
	function removeBlock(int $id): void{
		foreach($this->blocks as $k => $v){
			if($id == $k){
				$v->delete();
				$v = new we_newsletter_block();
				unset($this->blocks[$id]);
			}
		}
	}

	/**
	 * add new group to newsletter
	 *
	 */
	function addGroup(): void{
		$this->groups[] = new we_newsletter_group();
	}

	/**
	 * remove newsletter group
	 *
	 * @param int $group
	 */
	function removeGroup(int $group): void{
		$link = $group + 1;
		foreach($this->blocks as $bk => $block){
			$arr = $block->getGroups();
			foreach($arr as $k => $v){
				if($v == $link){
					$arr[$k] = -1;
				}
				if($v > $link){
					$arr[$k] = $v - 1;
				}
			}
			foreach($arr as $k => $v){
				if($v == -1){
					unset($arr[$k]);
				}
			}
			$this->blocks[$bk]->setGroups($arr);
		}
		unset($this->groups[$group]);
	}

	/**
	 * check email syntax
	 *
	 * @throws we_exception_newsletter
	 */
	function checkEmails(): void{
		if(!we_check_email($this->Sender)){
			throw new we_exception_newsletter($this->Sender, self::MALFORMED_SENDER);
		}
		if(!we_check_email($this->Reply)){
			throw new we_exception_newsletter($this->Reply, self::MALFORMED_REPLY);
		}
		if(!we_check_email($this->Test)){
			throw new we_exception_newsletter($this->Test, self::MALFORMED_TEST);
		}

		foreach($this->groups as $v){
			$v->checkEmails(); //throws exception
		}
	}

	/**
	 * set log in db
	 *
	 * @param string $log
	 * @param string $param
	 */
	function addLog(string $log, string $param = ''): void{
		$this->db->query('INSERT INTO ' . NEWSLETTER_LOG_TABLE . ' SET ' . we_database_base::arraySetter([
				'NewsletterID' => $this->ID,
				'Log' => $log,
				'Param' => $param
		]));
	}

	/**
	 * clear log in db
	 *
	 */
	function clearLog(): void{
		$this->db->query('DELETE FROM ' . NEWSLETTER_LOG_TABLE . ' WHERE NewsletterID=' . intval($this->ID));
	}

	/**
	 * checks recursive the parents to detect if path is ok or not
	 *
	 * @param int $id
	 * @return bool
	 */
	function checkParents(int $id): bool{
		$count = 0;
		while($id > 0){
			if($count > 1000){
				break;
			}
			if($id == $this->ID){
				return false;
			}
			$h = getHash('SELECT IsFolder,ParentID FROM ' . NEWSLETTER_TABLE . ' WHERE ID=' . $id);
			if($h['IsFolder'] != 1){
				return false;
			}
			$id = $h['ParentID'];
			$count++;
		}
		return true;
	}

	/**
	 * fix childs path if parets changes
	 *
	 */
	function fixChildsPaths(): void{
		$oldpath = fS('SELECT Path FROM ' . NEWSLETTER_TABLE . ' WHERE ID=' . intval($this->ID));

		if(trim($oldpath) != '' && trim($oldpath) != '/'){
			$this->db->query('UPDATE ' . NEWSLETTER_TABLE . ' SET Path=REPLACE(Path,"' . $this->db->escape($oldpath) . '","' . $this->db->escape($this->Path) . '") WHERE Path LIKE "' . $this->db->escape($oldpath) . '%"');
		}
	}

	/**
	 * checks if filename is well formated
	 *
	 * @return bool
	 */
	function filenameNotValid(): bool{
		return preg_match('|[^a-z0-9\ \._\-]|i', $this->Text) ? true : false;
	}

	public static function getJSLangConsts(): string{
		return we_base_util::setLangString([
				'add_email' => g_l('modules_newsletter', '[add_email]'),
				'all_list' => g_l('modules_newsletter', '[all_list]'),
				'ask_to_preserve' => g_l('modules_newsletter', '[ask_to_preserve]'),
				'continue_camp' => g_l('modules_newsletter', '[continue_camp]'),
				'del_email_file' => (g_l('modules_newsletter', '[del_email_file]')),
				'delete_group_question' => g_l('modules_newsletter', '[delete_group_question]'),
				'delete_question' => g_l('modules_newsletter', '[delete_question]'),
				'edit_email' => g_l('modules_newsletter', '[edit_email]'),
				'email_delete' => g_l('modules_newsletter', '[email_delete]'),
				'email_delete_all' => g_l('modules_newsletter', '[email_delete_all]'),
				'email_exists' => (g_l('modules_newsletter', '[email_exists]')),
				'email_max_len' => (g_l('modules_newsletter', '[email_max_len]')),
				'empty_name' => (g_l('modules_newsletter', '[empty_name]')),
				'mailing_list' => g_l('modules_newsletter', '[mailing_list]'),
				'must_save' => (g_l('modules_newsletter', '[must_save]')),
				'must_save_preview' => (g_l('modules_newsletter', '[must_save_preview]')),
				'no_email' => (g_l('modules_newsletter', '[no_email]')),
				'no_newsletter_selected' => (g_l('modules_newsletter', '[no_newsletter_selected]')),
				'no_subject' => g_l('modules_newsletter', '[no_subject]'),
				'nothing_to_delete' => (g_l('modules_newsletter', '[nothing_to_delete]')),
				'nothing_to_save' => (g_l('modules_newsletter', '[nothing_to_save]')),
				'save_changed_newsletter' => g_l('modules_newsletter', '[save_changed_newsletter]'),
				'search_finished' => g_l('modules_newsletter', '[search_finished]'),
				'search_text' => g_l('modules_newsletter', '[search_text]'),
				'send_question' => g_l('modules_newsletter', '[send_question]'),
				'send_test_question' => g_l('modules_newsletter', '[send_test_question]'),
				'test_email_question' => g_l('modules_newsletter', '[test_email_question]'),
				'retry' => g_l('modules_newsletter', '[retry]'),
		]);
	}

}
