<?php

declare(strict_types=1);

/* webEdition CMS
 *
 * $Rev: dbe3ba5afdec $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-03-04 13:06 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */

abstract class we_tag_var extends we_weTag_tag{

	public const BDID_TYPES = ['collection', 'customer', 'img', 'object', 'video'];
	public const UNKNOWN_TYPES = ['select', 'txt'];

	/**
	 * @param array<string,mixed> $attribs parsed tag-attributes
	 * @param string $content content between openingtag and closing tag @unused-param
	 * @param bool $internal if used by we internally @unused-param
	 * @return string output to print
	 * @throws we_exception_parser if any attribute is missing
	 */
	public static function tag(array $attribs = [], string $content = '', bool $internal = false): string{
		switch(self::getAttribute('type', $attribs, '', we_base_request::STRING)){ //Fix #9311
			case 'shopVat':
			case 'shopCategory': //shopVat and shopCategory need no attribute 'name'
				break;
			default:
				self::attributeMissing($attribs, ['name' => false], __CLASS__);
		}
		self::$doc = we_base_WE::getWeDoc();

		$docAttr = self::getAttribute('doc', $attribs, '', we_base_request::STRING);
		$name = self::getAttribute('name', $attribs, '', we_base_request::STRING);
		$name_orig = self::getAttribute('_name_orig', $attribs, '', we_base_request::STRING);
		$type = self::getAttribute('type', $attribs, '', we_base_request::STRING);
		$htmlspecialchars = self::getAttribute('htmlspecialchars', $attribs, false, we_base_request::BOOL); // #3771
		$format = self::getAttribute('format', $attribs, '', we_base_request::STRING);
		$num_format = self::getAttribute('num_format', $attribs, '', we_base_request::STRING);
		$doc = self::getDocForTag($docAttr, false);
		$varType = self::getAttribute('varType', $attribs, we_base_request::STRING, we_base_request::STRING);
		$prepareSQL = self::getAttribute('prepareSQL', $attribs, false, we_base_request::BOOL);
		$attribs = self::removeAttribs($attribs, ['varType', 'prepareSQL', 'htmlspecialchars',]);

		switch($type){
			case 'session' :
				$return = (string) self::getArrayValue($_SESSION, null, $name_orig);
				break;
			case 'cookie':
				$return = we_base_util::rmPhp((string) we_base_request::filterVar(self::getArrayValue($_COOKIE, null, $name_orig), $varType)); //@phan-suppress-current-line PhanDeprecatedFunction
				break;
			case 'request' :
				$return = we_base_util::rmPhp((string) we_base_request::filterVar(self::getArrayValue($_REQUEST, null, $name_orig), $varType)); //@phan-suppress-current-line PhanDeprecatedFunction
				break;
			case 'post' :
				$return = we_base_util::rmPhp((string) we_base_request::filterVar(self::getArrayValue($_POST, null, $name_orig), $varType)); //@phan-suppress-current-line PhanDeprecatedFunction
				break;
			case 'get' :
				$return = we_base_util::rmPhp((string) we_base_request::filterVar(self::getArrayValue($_GET, null, $name_orig), $varType)); //@phan-suppress-current-line PhanDeprecatedFunction
				break;
			case 'global':
				$return = (string) self::getArrayValue($GLOBALS, null, $name_orig);
				break;
			case 'multiobject' :
				$data = we_unserialize((string) $doc->getField($attribs, $type, true));
				$return = is_array($data) ? (!empty($data['objects']) ? implode(',', $data['objects']) : implode(',', $data)) : '';
			case 'property' :
				try{
					$return = we_base_WE::getWeObj()->$name_orig;
				}catch(we_exception_base $ex){
					$return = $doc->$name_orig;
				}
				break;
			case 'shopVat' :
				if(!we_modules_info::isActive(we_modules_info::SHOP)){
					return '';
				}
				if(!we_shop_category::isCategoryMode()){
					$return = (string) we_shop_vats::getVatRateForSite($doc->getElementI(we_shop_shop::VAT_FIELD_NAME));
				} else {
					$attribs['shopcategoryid'] = $doc->getElementI(we_shop_shop::CATEGORY_FIELD_NAME);
					$attribs['wedoccategories'] = $doc->Category;
					unset($attribs['type']);
					$return = (string) we_weTag_util::tag('shopVat', $attribs) ?: '';
				}
				break;
			case 'shopCategory' :
				$return = (string) (we_modules_info::isActive(we_modules_info::SHOP) ?
						$doc->getElementI(we_shop_shop::CATEGORY_FIELD_NAME) :
						'');
				break;
			case 'link' :
				$return = (string) $doc->getField($attribs, 'link', false);
				break;
			// bugfix #3634
			default:
				$return = self::getDefaultHandling($attribs, $type, $name_orig, $name, $doc);
		}
		if($return instanceof DateTime){
			return $return->format($format ?: g_l('date', '[format][php]'));
		}

		if($format && $type !== 'date'){//date
			return date($format, (int) $return);
		}
		if($num_format){
			return we_base_util::formatNumber((float) $return, $num_format);
		}
		$return = $htmlspecialchars ? htmlspecialchars((string) $return) : (string) $return;
		return ($prepareSQL ? DB_WE::$inst->escape($return) : $return);
	}

	private static function getDefaultHandling(array $attribs, string $type, string $name_orig, string $name, we_document_base $doc): string{
		$normVal = (string) $doc->getField($attribs, $type, true);
		$normVal = we_wysiwyg_editor::parseGalleryTags($normVal);

		// bugfix 7557
		// wenn die Abfrage im Aktuellen Objekt kein Ergebnis liefert
		// wird in den eingebundenen Objekten ueberprueft ob das Feld existiert
		$name = ($type === 'select' && $normVal == '' ? $name_orig : $name);
		$selectKey = self::getAttribute('key', $attribs, false, we_base_request::BOOL);
		if($type === 'select' && $selectKey){
			return $doc->getElementS($name);
		}

		if($doc instanceof we_contents_objectFile){
			$keys = array_keys($doc->DefArray);
			foreach($keys as $glob_key){
				$key = (string) $glob_key;
				if(str_starts_with($key, we_contents_object::QUERY_PREFIX)){
					$rest = substr($key, strlen(we_contents_object::QUERY_PREFIX));
				} elseif(str_starts_with($key, we_listview_base::OBJECT_FIELD_PREFIX)){
					$rest = substr($key, strlen(we_listview_base::OBJECT_FIELD_PREFIX));
				} else {
					$rest = '';
				}
				if($rest){
					$normVal = $doc->getFieldByVal($doc->getElementS($name), $type, $attribs, false, self::$doc->ParentID, self::$doc->Path, (int) $rest);
				}

				if($normVal != ''){
					break;
				}
			}
		}
		// EOF bugfix 7557
		return (string) $normVal;
	}

	public static function getElementData(we_document_webEdition $doc, string $name){
		$elemType = $doc->getElementType($name);
		return (in_array($elemType, self::UNKNOWN_TYPES) ?
				($doc->getElementS($name) ?: $doc->getElementI($name)) :
				(in_array($elemType, self::BDID_TYPES) ?
				$doc->getElementI($name) :
				$doc->getElementS($name)
				)
				);
	}

}
