<?php

declare(strict_types=1);

/**
 * webEdition CMS
 *
 * $Rev: 7bf7d093213c $
 * $Author: Marc Kraemer <marc.kraemer@webedition.org> $
 * $Date: 2021-02-04 17:15 +0100 $
 *
 * This source is part of webEdition CMS. webEdition CMS is
 * free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 3 of the License, or
 * any later version.
 *
 * The GNU General Public License can be found at
 * http://www.gnu.org/copyleft/gpl.html.
 * A copy is found in the textfile
 * webEdition/licenses/webEditionCMS/License.txt
 *
 * @license	http://www.gnu.org/copyleft/gpl.html  GPL
 */
class we_workflow_document extends we_workflow_base{

	const STATUS_ACTIVE = 'active';
	const STATUS_FINISHED = 'finished';
	const STATUS_CANCELED = 'canceled';

	//properties

	/** @var int */
	var $ID = 0;

	/** @var int */
	var $workflowID = 0;

	/** @var int */
	var $documentID = 0;

	/** @var string */
	protected $DocumentTable = "tblFile";

	/** @var int */
	var $userID = 0;

	/** @var string */
	var $Status = self::STATUS_ACTIVE;

	/** @var we_workflow_workflow */
	var $workflow;

	/** @var we_document_base */
	var $document;

	/** @var we_workflow_documentStep[] */
	var $steps = [];

	/**
	 * Default Constructor
	 *
	 */
	public function __construct(int $id = 0){
		parent::__construct(WORKFLOW_DOC_TABLE);
		$this->persistent_slots = [
			'ID' => we_base_request::INT,
			'workflowID' => we_base_request::INT,
			'documentID' => we_base_request::INT,
			'DocumentTable' => we_base_request::STRING,
			'userID' => we_base_request::INT,
			'Status' => we_base_request::STRING,
		];

		$this->load($id);
	}

	/**
	 * Load data from database
	 * @param bool $loadData @unused-param
	 */
	function load(int $id = 0, bool $loadData = true): bool{
		if($id){
			$this->ID = $id;
		}

		if(!$this->ID){
			return false;
		}
		if(!parent::load()){
			return false;
		}

		$this->workflow = new we_workflow_workflow($this->workflowID);

		$docTable = ($this->workflow->Type === we_workflow_workflow::OBJECT ? OBJECT_FILES_TABLE : FILE_TABLE);
		$this->DocumentTable = stripTblPrefix($docTable);
		if(($ct = fS('SELECT ContentType FROM ' . $docTable . ' WHERE ID=' . intval($this->documentID)))){
			$this->document = we_base_ContentTypes::getObject($ct);
			assert($this->document instanceof we_document_base);

			$this->document->initByID($this->documentID, '', we_contents_base::LOAD_TEMP_DB);
		}

		$this->steps = we_workflow_documentStep::getAllSteps($this->ID);
		return true;
	}

	function approve(int $uID, string $desc, bool $force = false): bool{
		$i = $this->findLastActiveStep();
		$ret = $this->steps[$i]->approve($uID, $desc, $force);
		if($this->steps[$i]->Status === we_workflow_documentStep::STATUS_APPROVED){
			$this->nextStep($i, $desc, $uID);
		}
		return $ret;
	}

	function autopublish(int $uID, bool $force = false): bool{
		$i = $this->findLastActiveStep();
		$ret = $this->steps[$i]->approve($uID, '', $force);
		if($this->steps[$i]->Status === we_workflow_documentStep::STATUS_APPROVED){
			$this->finishWorkflow(true, $uID);
			$this->document->save();
			try{
				$this->document->publInScheduleTable();
			}catch(we_exception_base $e){
				$this->document->we_publish();
			}

			$this->sendMail($this->userID, g_l('modules_workflow', '[auto_published]') . ($this->workflow->EmailPath ? ' ' . $this->document->Path : ''), g_l('modules_workflow', '[auto_published]') . "\n\n" . $this->document->Path);
		}
		return $ret;
	}

	function decline(int $uID, string $desc, bool $force = false): bool{
		$i = $this->findLastActiveStep();
		$ret = $this->steps[$i]->decline($uID, $desc, $force);
		if($this->steps[$i]->Status === we_workflow_documentStep::STATUS_CANCELED){
			$this->finishWorkflow(true, $uID);
			$desc = str_replace('<br/>', "\n", $desc);
			$this->sendMail($this->userID, g_l('modules_workflow', '[todo_returned]') . ($this->workflow->EmailPath ? ' ' . $this->document->Path : ''), g_l('modules_workflow', '[todo_returned]') . "\n\n" . $desc . "\n\n" . $this->document->Path);
		}
		return $ret;
	}

	private function nextStep(int $index = -1, string $desc = '', int $uid = 0): void{
		if($index < 0){
			return;
		}
		if($index < count($this->steps) - 1){
			$this->steps[$index + 1]->start($desc);
		} else {
			$this->finishWorkflow(false, $uid);
		}
	}

	function finishWorkflow(bool $force = false, int $uID = 0): bool{
		if($force){
			$this->Status = self::STATUS_CANCELED;
			foreach($this->steps as &$sv){
				if($sv->Status === we_workflow_documentStep::STATUS_NEW){
					$sv->Status = we_workflow_documentStep::STATUS_CANCELED;
				}
				foreach($sv->tasks as &$tv){
					if($tv->Status === we_workflow_documentTask::STATUS_NEW){
						$tv->Status = we_workflow_documentTask::STATUS_CANCELED;
					}
				}
			}
			//insert into document Log
			we_workflow_log::logDocumentEvent($this->ID, $uID, we_workflow_log::TYPE_DOC_FINISHED_FORCE);
			return true;
		}
		$this->Status = self::STATUS_FINISHED;
		we_workflow_log::logDocumentEvent($this->ID, $uID, we_workflow_log::TYPE_DOC_FINISHED);

		return true;
	}

	/**
	 * Create next step or finish workflow document if last step is done
	 *
	 */
	/* function createNextStep(string $stepKey, $uid = 0){
	  if($stepKey >= count($this->steps)){
	  // no more steps, finish workflow
	  return $this->finishWorkflow(0, $uid);
	  }
	  $step = &$this->steps[$stepKey];
	  $step->start();
	  return true;
	  } */

	/**
	 * Find last document Status step
	 *
	 */
	function findLastActiveStep(): int{
		for($i = count($this->steps) - 1; $i >= 0; $i--){
			if($this->steps[$i]->hasStarted()){
				return $i;
			}
		}
		throw new we_exception_base('not found');
	}

	/** save workflow document in database
	 *
	 * @param bool $force_new @unused-param
	 * @param bool $rebuild @unused-param
	 * @param array $skipWrite @unused-param
	 * @return bool
	 */
	function save(bool $force_new = false, bool $rebuild = false, array $skipWrite = []): bool{
		if(!$this->documentID){
			return false;
		}
		parent::save();
		foreach($this->steps as $cur){
			$cur->setDocID($this->ID);
			$cur->save();
		}
		return true;
	}

	function delete(): bool{
		if(!$this->ID){
			return false;
		}

		foreach($this->steps as $v){
			$v->delete();
		}
		parent::delete();
		return true;
	}

	/*	 * ***************** STATIC FUNCTIONS**************************
	  /**
	 * return workflowDocument for document
	 *    return false if no workflow
	 */

	public static function find(int $documentID, array $type, string $status): ?self{
		$id = fI('SELECT doc.ID FROM ' .
				WORKFLOW_DOC_TABLE . ' doc JOIN ' .
				WORKFLOW_TABLE . ' wf ON doc.workflowID=wf.ID
WHERE
	doc.documentID=' . $documentID . ' AND
	doc.Status="' . DB_WE::$inst->escape($status) . '"' .
				($type ? ' AND wf.Type IN ("' . implode('","', $type) . '")' : '') . '
ORDER BY
	doc.ID DESC LIMIT 1');
		return ($id ? new self($id) : null);
	}

	/**
	 * Create new workflow document
	 *    if workflow for that document exists, function will return it
	 */
	public static function createNew(int $documentID, array $type, int $workflowID, int $userID): self{
		if(($newWfDoc = self::find($documentID, $type, self::STATUS_ACTIVE))){
			return $newWfDoc;
		}

		//fixme: check the difference to new self($documentID)

		$newWFDoc = new self();
		$newWFDoc->documentID = $documentID;
		$newWFDoc->userID = $userID;
		$newWFDoc->workflowID = $workflowID;
		$newWFDoc->workflow = new we_workflow_workflow($workflowID);
		$newWFDoc->steps = we_workflow_documentStep::createAllSteps($workflowID);

		return $newWFDoc;
	}

}
